#!/usr/bin/env python3
'''
# Copyright (C) 2020, Elphel.inc.
# Usage: known
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http:#www.gnu.org/licenses/>.

@author:     Oleg K Dzhimiev, Konstantyn Chebanov
@copyright:  2016 Elphel, Inc.
@license:    GPLv3.0+
@contact:    oleg@elphel.com
@deffield    updated: unknown

'''

__author__ = "Elphel"
__copyright__ = "Copyright 2020, Elphel, Inc."
__license__ = "GPL"
__version__ = "4.0"
__maintainer__ = "Oleg K Dzhimiev"
__email__ = "oleg@elphel.com"
__status__ = "Development"

import subprocess
import sys
import os
import time

# functions
# useful link 1: http://superuser.com/questions/868117/layouting-a-disk-image-and-copying-files-into-it
# useful link 2: poky/scripts/contrib/mkefidisk.sh
# (?) useful link 3: http://unix.stackexchange.com/questions/53890/partitioning-disk-image-file

def shout(cmd):
    """Execute shell command and print to console"""
    subprocess.call(cmd, shell=True)

def print_help():
    """Print help information"""
    print("\nDescription:\n")
    print("  * Required programs: kpartx, parted")
    print("  * Run under superuser. Make sure the correct device is provided.")
    print("  * Erases partition table on the provided device")
    print("  * If given someimage.img file - burns the sd card from it")
    print("  * If not - uses the files from the predefined list")
    print("  * Creates FAT32 partition labeled 'BOOT' and copies files required for boot")
    print("  * Creates EXT4 partition labeled 'root' and extracts rootfs.tar.gz")
    print("\nExamples:\n")
    print("  * Use files (names are hardcoded) from the current dir ('build/tmp/deploy/images/elphel393/mmc/'):")
    print("      ~$ sudo python3 make_sdcard.py /dev/sdz")
    print("  * Use someimage.img file:")
    print("      ~$ sudo python3 make_sdcard.py /dev/sdz someimage.img")
    print("  * To write *.iso use a standard OS tool that burns bootable USB drives")
    print("")

def check_program_installed(program):
    """Check if a program is installed"""
    try:
        result = subprocess.check_output(["which", program], stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        print(f"Program missing: {program}")
        return False

# Check required programs
required_programs = (
    "parted",
    "kpartx"
)

something_is_missing = False
for program in required_programs:
    if not check_program_installed(program):
        something_is_missing = True

if something_is_missing:
    print("\nPlease install missing programs and run again.")
    sys.exit(1)

# Parse command line arguments
if len(sys.argv) > 1:
    DEVICE = sys.argv[1]
else:
    DEVICE = ""
    print_help()
    sys.exit(0)

if len(sys.argv) > 2:
    IMAGE_FILE = sys.argv[2]
    if not IMAGE_FILE.endswith(".img"):
        print("ERROR: Please provide *.img file or leave argument empty to use certain image files in the current dir")
        sys.exit(1)
else:
    IMAGE_FILE = ""

print("NOTE: If plasma crashes, do not worry")

# Parameters
SDCARD_SIZE = 4000

PT_TYPE = "msdos"

BOOT_LABEL = "BOOT"
BOOT_FS = "fat32"
BOOT_SIZE = 128
BOOT_FILE_LIST = (
    "boot.bin",
    "u-boot-dtb.img",
    "devicetree.dtb",
    "uImage"
)

ROOT_LABEL = "root"
ROOT_FS = "ext4"
ROOT_ARCHIVE = "rootfs.tar.gz"

something_is_missing = False

# Check if required files exist
if IMAGE_FILE == "":
    print(f"Preparing SD card using files: {BOOT_FILE_LIST + (ROOT_ARCHIVE,)}")
    for f in BOOT_FILE_LIST + (ROOT_ARCHIVE,):
        if not os.path.isfile(f):
            print(f"File {f} is missing")
            something_is_missing = True
else:
    print(f"Preparing SD card from {IMAGE_FILE}")
    if not os.path.isfile(IMAGE_FILE):
        print("No such file")
        something_is_missing = True

if not os.path.exists(DEVICE):
    print(f"No such device: {DEVICE}")
    something_is_missing = True

if something_is_missing:
    sys.exit(1)

# Main execution
print(f"= Erase partition table on {DEVICE} and everything else")
shout(f"dd if=/dev/zero of={DEVICE} bs=512 count=2")
shout(f"dd if=/dev/zero of={DEVICE} bs=1MB count=1 seek=1")
shout(f"dd if=/dev/zero of={DEVICE} bs=1MB count=1 seek={BOOT_SIZE}")

print("= Create partition table")
shout(f"parted -s {DEVICE} mktable {PT_TYPE}")

print("= Create partitions")
shout(f"parted -s {DEVICE} mkpart primary {BOOT_FS} 1 {BOOT_SIZE}")
shout(f"parted -s {DEVICE} mkpart primary {ROOT_FS} {BOOT_SIZE + 1} 100%")
# Check alignment
shout(f"parted -s {DEVICE} align-check optimal 1")
shout(f"parted -s {DEVICE} align-check optimal 2")

# Wait for device nodes to be created
devs_created = False
partition1 = f"{DEVICE}1" if not "mmcblk" in DEVICE else f"{DEVICE}p1"
partition2 = f"{DEVICE}2" if not "mmcblk" in DEVICE else f"{DEVICE}p2"

print("= Waiting for device nodes...")
while not devs_created:
    if os.path.exists(partition1) and os.path.exists(partition2):
        devs_created = True
    else:
        print(".", end="", flush=True)
        time.sleep(0.5)
print()

time.sleep(1)
print("= Format partitions")

shout(f"mkfs.vfat {partition1} -F 32 -n {BOOT_LABEL}")
shout(f"mkfs.ext4 {partition2} -F -L {ROOT_LABEL}")

# Create temporary mount point
shout("mkdir -p tmp")

if IMAGE_FILE == "":
    # Copy boot files
    print("= Copying boot files...")
    shout(f"mount {partition1} tmp")
    for i in BOOT_FILE_LIST:
        print(f"    {i}")
        shout(f"cp {i} tmp")
    shout("umount tmp")

    # Extract root filesystem
    print("= Extracting root filesystem...")
    shout(f"mount {partition2} tmp")
    shout(f"tar -C tmp/ -xzpf {ROOT_ARCHIVE}")
    shout("umount tmp")
else:
    # Copy from image file
    print("= Copying from image file...")
    shout("modprobe dm-mod")
    shout(f"kpartx -av {IMAGE_FILE}")
    
    # Wait for mapper devices
    devs_created = False
    print("= Waiting for mapper devices...")
    while not devs_created:
        if os.path.exists("/dev/mapper/loop0p1") and os.path.exists("/dev/mapper/loop0p2"):
            devs_created = True
        else:
            print(".", end="", flush=True)
            time.sleep(0.5)
    print()
    
    shout("mkdir -p tmp2")
    
    # Copy boot partition
    print("= Copying boot partition...")
    shout(f"mount {partition1} tmp")
    shout("mount /dev/mapper/loop0p1 tmp2")
    shout("rsync -a tmp2/ tmp")
    shout("umount tmp")
    shout("umount tmp2")
    
    # Copy root partition
    print("= Copying root partition...")
    shout(f"mount {partition2} tmp")
    shout("mount /dev/mapper/loop0p2 tmp2")
    shout("rsync -a tmp2/ tmp")    
    shout("umount tmp")
    shout("umount tmp2")
    
    shout("rm -rf tmp2")
    shout(f"kpartx -dv {IMAGE_FILE}")

# Clean up
shout("rm -rf tmp")

print("\nDone! SD card is ready for use.")
print(f"You can now safely remove {DEVICE}")
